#' Export a Formattable as PNG, PDF, or JPEG
#'
#' @param f A formattable.
#' @param file Export path with extension .png, .pdf, or .jpeg.
#' @param width Width specification of the html widget being exported.
#' @param height Height specification of the html widget being exported.
#' @param background Background color specification.
#' @param delay Time to wait before taking webshot, in seconds.
#'
#' @importFrom formattable as.htmlwidget
#' @importFrom htmltools html_print
#' @importFrom webshot webshot
#'
#' @export
export_formattable <- function(f, file, width = "100%", height = NULL, 
                               background = "white", delay = 0.2)
{
  w <- as.htmlwidget(f, width = width, height = height)
  path <- html_print(w, background = background, viewer = NULL)
  url <- paste0("file:///", gsub("\\\\", "/", normalizePath(path)))
  webshot(url,
          file = file,
          selector = ".formattable_widget",
          delay = delay)
}

###############################################################################
################### MULTIPLE HISTOGRAMS IN ONE FIGURE FUNCTION ################
###############################################################################

## Functions you'll need

plot_multi_histogram <- function(df, feature, label_column) {
  plt <- ggplot(df, aes(x = eval(parse(text=feature)), fill = eval(parse(text = label_column)))) +
    geom_histogram(alpha = 0.7, position="identity", 
                   aes(y = ..density..), color = "black") +
    geom_density(alpha = 0.7) +
    geom_vline(aes(xintercept = 0), color = "black", linetype = "dashed", size = 1) +
    labs(x = feature, y = "Density") + 
    theme_custom()    
  plt + guides(fill = guide_legend(title = label_column)) +
    ggtitle("Effect Size Distribution by Tax Type")
}


##################################################################
################### CREATE LATEX TABLE FROM DATAFRAME #############
###################################################################

## Shape should be coefficients in the rows, QOIs and model
## differentiating variable in the columns

prep_texreg <- function(data = NULL, 
                       names.col = "term", # Should be names of coefficients
                       coef.col = "estimate", 
                       p.col = "p.value", 
                       se.col = "std.error",
                       lci.col = "conf.low", 
                       uci.col = "conf.high",
                       model.col = "outcome", 
                       alpha = 0.05){ # Names of goodness of fit statistics
                       
  require(multiwayvcov)
  df <- data
  texreg_obj <- list() # This is where you store the texreg objects
  arg_names <- c(names.col, coef.col, p.col, se.col, 
                 lci.col, uci.col, model.col)
  arg_na <- arg_names[!arg_names %in% ls(df)]
  
  if(any(c(lci.col, uci.col) %in% arg_na)){
    df[, lci.col] <- df[, coef.col] - qnorm(1 - alpha/2)*df[, se.col]
    df[, uci.col] <- df[, coef.col] + qnorm(1 - alpha/2)*df[, se.col]
    warning("No confidence intervals supplied or default column 
    names do not match those in data frame. Creating symmetric confidence intervals from
            standard error.")}
  
  if(class(df[, model.col]) == "character")
    mod_names <- unique(df[, model.col])
  if(class(df[, model.col]) == "factor")
    mod_names <- levels(df[, model.col])
  
  for(i in 1:length(mod_names)){
    df_sub <- df[df[, model.col] == mod_names[i], ]
    
    if(any(c(lci.col, uci.col) %in% arg_na))
      texreg_obj[[i]] <- createTexreg(coef = df_sub[, coef.col], 
                                      pvalues = df_sub[, p.col], 
                                      coef.names = df_sub[, names.col], 
                                      se = df_sub[, se.col], 
                                      model.name = mod_names[i])
    
    if(!all(c(lci.col, uci.col) %in% arg_na))
      texreg_obj[[i]] <- createTexreg(coef = df_sub[, coef.col], 
                                      pvalues = df_sub[, p.col], 
                                      coef.names = df_sub[, names.col], 
                                      se = df_sub[, se.col], 
                                      ci.low = df_sub[, lci.col], 
                                      ci.up = df_sub[, uci.col], 
                                      model.name = mod_names[i])
    
      
    }
  
  return(list(obj = texreg_obj, 
              mod_names = mod_names))
  
}
  
###################################################################
################### CUSTOM GGPLOT FUNCTION ########################
###################################################################

theme_custom <- function(legend_justification = c(1, 0), 
                         legend_position = c(0.95, 0.8), 
                         legend_box = "vertical", 
                         title_size = 12,
                         title_position = 0.5, # 0 = left, 1 = right
                         subtitle_position = 0.5,
                         axis_text_size = 10, 
                         axis_title_size = 12,
                         axis_border_size = 0,
                         text_size = 12,
                         panel_border_size = 0.7){
  theme_bw() +
    theme(panel.grid.major = element_blank(), 
          panel.grid.minor = element_blank(), 
          panel.background = element_blank(),
          title = element_text(size = title_size),
          axis.text = element_text(size = axis_text_size),
          axis.title = element_text(size = axis_title_size),
          axis.line = element_line(size = axis_border_size),
          legend.justification = legend_justification, 
          legend.position = legend_position,
          legend.box = legend_box,
          panel.border = element_rect(colour = "black", fill = NA, size = panel_border_size), 
          panel.spacing = unit(0, "mm"), 
          plot.title = element_text(hjust = title_position), 
          plot.subtitle = element_text(hjust = subtitle_position))  
}

##################################################################################################
############### MAKE EQUATION AS CHARACTER STRING FROM DVs and IVs ###############################
##################################################################################################

make_equation <- function(dvs, ivs, interactions = c()){
  return(paste(dvs, "~", paste(c(ivs, interactions), sep = "", collapse = " + ")))
}


###################################################################
################### QUICK REGRESSION + VISUALIZATION ##############
###################################################################

quick_lm <- function(dep.vars = NULL, 
                     ind.vars = NULL, # note that the order these variables are in will control the order they appear in the plot
                     data = NULL, 
                     alpha = 0.05,
                     dep.names = NULL,
                     data.names = NULL,
                     stand.vars = NULL, # Set this to "non-binary" if you want all non-binary, non-factor/character class variables standardized
                     stand.mult = 2, # This is the multiplier for the standard deviation scaling on plots. Default is 2 to make it more comparable to binary predictors
                     inc.mod = F,
                     inc.plot = F,
                     print.plot = F,
                     keep.vars.tab = NULL, 
                     keep.dep.tab = NULL,
                     keep.data.tab = NULL,
                     keep.vars.plot = NULL, 
                     keep.dep.plot = NULL, 
                     keep.data.plot = NULL,
                     title.plot = c(""), 
                     x.size = 14, 
                     y.size = 12, 
                     strip.size = 14,
                     subset.by = "dataset", # Can be either "dataset" or "outcome"
                     ...){
  
  ################ PACKAGES AND CHECKS #########################
  
  packages.needed <- c("lmtest", "clubSandwich", 
                       "dotwhisker", "broom", "estimatr", 
                       "dplyr")
  if(!all(packages.needed %in% rownames(installed.packages()))){
    stop(paste("Need to install the following packages:",
               packages.needed[!packages.needed %in% rownames(installed.packages())]))}
  lapply(packages.needed, require, character.only = TRUE)
  
  if(class(data) != "list"){data <- list(data)}
  if(any(dep.vars %in% ind.vars)){stop("At least one dependent variable is also be included as a predictor")}
  if(is.null(dep.names)){dep.names <- dep.vars}
  if(is.null(data.names)){data.names <- paste("dataset", 1:length(data), sep = "")}
  if(is.null(subset.by) & length(dep.vars) == 1 & length(data) > 1){
    subset.by <- "outcome"}
  if(is.null(subset.by) & length(dep.vars) > 1 & length(data) == 1){
    subset.by <- "dataset"}
  if(is.null(keep.vars.tab)){keep.vars.tab <- c(ind.vars, "(Intercept)")}
  if(is.null(keep.dep.tab)){keep.dep.tab <- dep.vars}
  if(is.null(keep.data.tab)){keep.data.tab <- data.names}
  if(is.null(keep.vars.plot)){keep.vars.plot <- c(ind.vars, "(Intercept)")}
  if(is.null(keep.dep.plot)){keep.dep.plot <- dep.vars}
  if(is.null(keep.data.plot)){keep.data.plot <- data.names}
  
  ################ ESTIMATION ########################
  
  
  est.eq <- make_equation(dep.vars, ind.vars)
  data.tab <- info.tab <- c()
  get.stats <- c("N", "k", "r.squared", "adj.r.squared")
  mod.data <- list() # This will store model objects
  
  ## Loop over datasets
  
  for(i in 1:length(data)){
    
    ## Check to make sure variable is in each dataset   
    if(!all(c(ind.vars, dep.vars) %in% colnames(data[[i]]))){
      all.vars <- c(ind.vars, dep.vars)
      miss.vars <- all.vars[!all.vars %in% colnames(data[[i]])]
      if(length(miss.vars > 0)){
        stop(paste("The following variables are not present in", data.names[i], ":", 
                   paste(miss.vars, collapse = ", ")))
      }
    }
    
    ## Standardize according to stand.vars argument
    if(all(stand.vars == "non-binary") & !is.null(stand.vars)){
      all.vars <- c(ind.vars, dep.vars)
      change.vars <- all.vars[unlist(lapply(data[[i]][, all.vars], 
                                            FUN = function(x){
                                              ifelse(length(unique(x[!is.na(x)])) > 2 & class(x) %in% c("numeric", "integer"), T, F)}))]
      data[[i]][, change.vars] <- scale(data[[i]][, change.vars], 
                                        scale = stand.mult*apply(data[[i]][, change.vars], 2, sd, na.rm = T))
    }
    
    if(all(stand.vars != "non-binary") & !is.null(stand.vars)){
      all.vars <- c(ind.vars, dep.vars)
      add.vars <- stand.vars[!stand.vars %in% c(ind.vars, dep.vars)]
      if(length(add.vars > 0)){
        stop(paste("The following variables are set for standardization but are not in the dependent or independent variable vector", 
                   add.vars))}
      
      change.vars <- stand.vars[unlist(lapply(data[[i]][, stand.vars], 
                                              FUN = function(x){
                                                ifelse(length(unique(x[!is.na(x)])) > 2 & class(x) %in% c("numeric", "integer"), T, F)}))]
      data[[i]][, change.vars] <- scale(data[[i]][, change.vars], 
                                        scale = stand.mult*apply(data[[i]][, change.vars], 2, sd, na.rm = T))
    }
    
    ## Build necessary objects
    coef.tab <- stats.tab <- c()
    mod.dep <- list()
    ## Loop over dependent variables
    for(j in 1:length(dep.vars)){
      main.args <- list(formula = as.formula(est.eq[j]), 
                        data = data[[i]], 
                        alpha = alpha)
      add.args <- list(...)
      mod <- mod.dep[[j]] <- do.call(lm_robust, args = c(main.args, add.args))
      tmp.tab <- tidy(mod)
      stats.tab <- rbind(stats.tab, c(data.names[i], dep.names[j], 
                                      mod[get.stats], mod$fstatistic[1]))
      coef.tab <- rbind(coef.tab, tmp.tab)                  
      
    }
    names(mod.dep) <- dep.vars
    mod.data[[i]] <- mod.dep
    
    ## Clean up the info and tab objects a bit and stack them
    stats.tab <- as.data.frame(stats.tab)
    colnames(stats.tab) <- c("dataset", "outcome", get.stats, "fstatistic")
    info.tab <- rbind(info.tab, stats.tab)
    coef.tab$dataset <- data.names[i]
    data.tab <- rbind(data.tab, coef.tab)
  }
  names(mod.data) <- data.names
  
  colnames(data.tab) <- c("term", "estimate", "std.error", "statistic", 
                          "p.value", "conf.low", "conf.high", "df", 
                          "outcome", "dataset")
  data.tab <- data.tab[,  c("dataset", "outcome", "term", "estimate", "std.error", 
                            "statistic", "p.value", "conf.low", "conf.high")]
  data.tab$signif <- ifelse(data.tab$p.value <= alpha, 1, 0)
  
  ############## PLOTTING ###################
  
  if(inc.plot){
    
    plot.tab <- data.tab
    plot.tab$model <- factor(plot.tab$outcome, levels = dep.vars)
    keep.plot.rows <- Reduce(intersect, list(grep(paste(keep.vars.plot, collapse = "|"), plot.tab$term), 
                                             grep(paste(keep.dep.plot, collapse = "|"), plot.tab$outcome), 
                                             grep(paste(keep.data.plot, collapse = "|"), plot.tab$dataset)))
    plot.tab <- plot.tab[keep.plot.rows, ]
    levels(plot.tab$model) <- dep.names
    single.plot <- sum(length(dep.vars), length(data)) == 2 
    
    ### IF SINGLE DV AND SINGLE DATASET
    
    if(single.plot){
      tab.use <- plot.tab
      p <- dwplot(tab.use, intercept = F, 
                  dot_args = list(size = 5, pch = 20, 
                                  aes(colour = abs(statistic))), 
                  whisker_args = list(lwd = 1.5, 
                                      aes(colour = abs(statistic)))) + 
        scale_colour_gradient(low = "lightgray", high = "black")
      
      p <- p +  theme_bw(base_size = 12) + 
        geom_vline(xintercept = 0, linetype = "dashed") + 
        theme(axis.text.x = element_text(size = x.size), 
              axis.text.y = element_text(size = y.size),
              strip.text = element_text(size = strip.size), 
              legend.position = "none", 
              plot.margin = margin(0.75, 0.75, 0.75, 0.75, "cm")) + 
        ggtitle(paste("DV:", dep.vars))
      
      #################### LABEL CHANGES AND OTHER ADDITIONS ##################
      
      all.stand <- all(dep.vars %in% stand.vars)
      all.orig <- all(!dep.vars %in% stand.vars)
      some.mixed <- all.stand == F & all.orig == F
      
      if(all.stand){
        p <- p + xlab("\nChange in Dependent Variable (Standard Deviation Units)")}
      if(all.orig){
        p <- p + xlab("\nChange in Dependent Variable (Original Units)")}
      if(some.mixed){
        p <- p + xlab("\nChange in Dependent Variable (Mixed Units)")
        warning("Only some dependent variables are standardized: changing x-axis label to `Mixed Units`")}
      plot.list <- p
      
      if(print.plot){print(p)}# This will make sure plot shows automatically
      
    }
    
    ### IF MULTIPLE DVs OR MULTIPLE DATASETS (OR BOTH)
    
    if(!single.plot){
      
      ## If you want to produce a separate plot for each outcome
      
      if(subset.by == "outcome"){
        
        plot.list <- list()
        loop.vals <- unique(plot.tab$outcome)
        
        for(i in 1:length(loop.vals)){
          
          tab.use <- plot.tab[plot.tab$outcome == loop.vals[i], ]
          p <- dwplot(tab.use, intercept = F, 
                      dot_args = list(size = 5, pch = 20, 
                                      aes(colour = abs(statistic))), 
                      whisker_args = list(lwd = 1.5, 
                                          aes(colour = abs(statistic)))) + 
            scale_colour_gradient(low = "lightgray", high = "black") + 
            
            
            if(length(unique(tab.use$dataset)) > 1){
              p <- p + facet_wrap(~dataset, labeller = label_parsed)
            }
          
          p <- p +  theme_bw(base_size = 12) + 
            geom_vline(xintercept = 0, linetype = "dashed") + 
            theme(axis.text.x = element_text(size = x.size), 
                  axis.text.y = element_text(size = y.size),
                  strip.text = element_text(size = strip.size), 
                  legend.position = "none", 
                  plot.margin = margin(0.75, 0.75, 0.75, 0.75, "cm")) + 
            ggtitle(paste("Estimated Effects in ", loop.vals[i], "sample"))
          
          #################### LABEL CHANGES AND OTHER ADDITIONS ##################
          
          all.stand <- all(dep.vars %in% stand.vars)
          all.orig <- all(!dep.vars %in% stand.vars)
          some.mixed <- all.stand == F & all.orig == F
          
          if(all.stand){
            p <- p + xlab("\nChange in Dependent Variable (Standard Deviation Units)")}
          if(all.orig){
            p <- p + xlab("\nChange in Dependent Variable (Original Units)")}
          if(some.mixed){
            p <- p + xlab("\nChange in Dependent Variable (Mixed Units)")
            warning("Only some dependent variables are standardized: changing x-axis label to `Mixed Units`")}
          
          if(print.plot){print(p)}# This will make sure plot shows automatically
          
          plot.list[[i]] <- p
          
        } # close i loop
        
        names(plot.list) <- dep.names 
        
      } # close the multiple outcomes loop
      
      
      ## IF YOU WANT TO PRODUCE A SEPARATE PLOT FOR EACH DATASET
      
      if(subset.by == "dataset"){
        
        loop.vals <- unique(plot.tab$dataset)
        plot.list <- list()
        
        for(i in 1:length(loop.vals)){
          
          tab.use <- plot.tab[plot.tab$dataset == loop.vals[i], ]
          p <- dwplot(tab.use, intercept = F, 
                      dot_args = list(size = 5, pch = 20, 
                                      aes(colour = abs(statistic))), 
                      whisker_args = list(lwd = 1.5, 
                                          aes(colour = abs(statistic)))) + 
            scale_colour_gradient(low = "lightgray", high = "black")
          
          if(length(unique(tab.use$model)) > 1){ # Note that have to use model here and not "outcome" because model is what is taken by dwplot
            p <- p + facet_wrap(~model, labeller = label_parsed)
          }
          
          p <- p +  theme_bw(base_size = 12) + 
            geom_vline(xintercept = 0, linetype = "dashed") + 
            theme(axis.text.x = element_text(size = x.size), 
                  axis.text.y = element_text(size = y.size),
                  strip.text = element_text(size = strip.size), 
                  legend.position = "none", 
                  strip.background = element_rect(size = 0.5), 
                  plot.margin = margin(0.75, 0.75, 0.75, 0.75, "cm")) + 
            ggtitle(paste("Results for", loop.vals[i]))
          
          #################### LABEL CHANGES AND OTHER ADDITIONS ##################
          
          all.stand <- all(dep.vars %in% stand.vars)
          all.orig <- all(!dep.vars %in% stand.vars)
          some.mixed <- all.stand == F & all.orig == F
          
          if(all.stand){
            p <- p + xlab("\nEstimated Effect (DV in Standard Deviation Units)")}
          if(all.orig){
            p <- p + xlab("\nEsimated Effect (DV in Original Units)")}
          if(some.mixed){
            p <- p + xlab("\nEstimated Effect (DV in Mixed Units)")
            warning("Only some dependent variables are standardized: changing x-axis label to `Mixed Units`")}
          
          if(print.plot){print(p)}# This will make sure plot shows automatically
          
          plot.list[[i]] <- p
          
        } # close i loop
        
        names(plot.list) <- data.names 
        
      } # Close the data subset loop
    } # Close the multiple plot loop
    
    
  } # Close plot include loop
  
  
  ############# SUBSET ACCORDING TO KEEP ARGUMENTS AND BUILD FINAL OBJECT
  
  keep.tab.rows <- Reduce(intersect, list(grep(paste(keep.vars.tab, collapse = "|"), data.tab$term), 
                                          grep(paste(keep.dep.tab, collapse = "|"), data.tab$outcome), 
                                          grep(paste(keep.data.tab, collapse = "|"), data.tab$dataset)))
  data.tab <- data.tab[keep.tab.rows, ]
  data.tab$outcome <- factor(data.tab$outcome, levels = dep.vars)
  levels(data.tab$outcome) <- dep.names
  
  if(!is.null(stand.vars)){print(paste("The following variables have been standardized:", 
                                       paste(change.vars, collapse = ", ")))}
  if(exists("plot.list")){print(paste("NOTE: Coefficient scaling on plots is", stand.mult))}
  if(!exists("plot.list")){plot.list <- "Set inc.plot to TRUE to include plot objects in output"}
  if(!inc.mod){mod.data <- "Set inc.mod to TRUE to include model objects in output"}
  if(!all(dep.vars == dep.names)){warning("Custom dependent variable names used, check that order in dep.names matches order in dep.vars argument")}
  return(list(table = as_tibble(data.tab),
              info = info.tab,
              models = mod.data,
              plot = plot.list))
  
}
